"use strict";(self.webpackChunkschedula_form=self.webpackChunkschedula_form||[]).push([[746],{1746:function(e,t,u){u.r(t),u.d(t,{default:function(){return n.Z},getTableFooterUtilityClass:function(){return r.c},tableFooterClasses:function(){return r.Z}});var n=u(6005),r=u(4935)}}]);