.flexible-modal{background:#fff;border:none;position:absolute;z-index:10000000000}.flexible-modal-mask{background:rgba(55,55,55,.6);bottom:0;display:none;height:100%;left:0;position:fixed;right:0;top:0}.flexible-modal-resizer{border-bottom:2px solid #333;border-right:2px solid #333;bottom:0;cursor:se-resize;margin:5px;position:absolute;right:0}.flexible-modal-drag-area{background:transparent none repeat 0 0/auto auto padding-box border-box scroll!important;background:initial!important;cursor:move;height:64px!important;left:0!important;position:absolute;right:64px!important;top:0;width:auto!important}.flexible-modal-drag-area-bottom,.flexible-modal-drag-area-left,.flexible-modal-drag-area-right,.flexible-modal-drag-area-top{display:none}body.reflex-col-resize{cursor:col-resize}body.reflex-row-resize{cursor:row-resize}.reflex-container{align-content:stretch;align-items:stretch;display:flex;height:100%;justify-content:flex-start;position:relative;width:100%}.reflex-container.horizontal{flex-direction:column;min-height:1px}.reflex-container.vertical{flex-direction:row;min-width:1px}.reflex-container>.reflex-element{height:100%;overflow:auto;position:relative;width:100%}.reflex-container.reflex-resizing>.reflex-element{pointer-events:none;-webkit-user-select:none;user-select:none}.reflex-container>.reflex-element>.reflex-size-aware{height:100%;width:100%}.reflex-container>.reflex-splitter{background-color:#eee;z-index:100}.reflex-container>.reflex-splitter.active,.reflex-container>.reflex-splitter:hover{background-color:#c6c6c6;transition:all 1s ease}.horizontal>.reflex-splitter{border-bottom:1px solid #c6c6c6;border-top:1px solid #c6c6c6;cursor:row-resize;height:2px;width:100%}.reflex-element.horizontal .reflex-handle{cursor:row-resize;-webkit-user-select:none;user-select:none}.reflex-container.horizontal>.reflex-splitter.active,.reflex-container.horizontal>.reflex-splitter:hover{border-bottom:1px solid #eee;border-top:1px solid #eee}.reflex-container.vertical>.reflex-splitter{border-left:1px solid #c6c6c6;border-right:1px solid #c6c6c6;cursor:col-resize;height:100%;width:2px}.reflex-element.vertical .reflex-handle{cursor:col-resize;-webkit-user-select:none;user-select:none}.reflex-container.vertical>.reflex-splitter.active,.reflex-container.vertical>.reflex-splitter:hover{border-left:1px solid #eee;border-right:1px solid #eee}.reflex-container>.reflex-splitter.reflex-thin{-moz-background-clip:padding;-webkit-background-clip:padding;background-clip:padding-box;box-sizing:border-box;opacity:.2;z-index:100}.reflex-container>.reflex-splitter.reflex-thin.active
.reflex-container>.reflex-splitter.reflex-thin:hover{opacity:.5;transition:all 1.5s ease}.reflex-container.horizontal>.reflex-splitter.reflex-thin{border-bottom:8px solid hsla(0,0%,100%,0);border-top:8px solid hsla(0,0%,100%,0);cursor:row-resize;height:17px!important;margin:-8px 0;width:100%}.reflex-container.horizontal>.reflex-splitter.reflex-thin.active,.reflex-container.horizontal>.reflex-splitter.reflex-thin:hover{border-bottom:8px solid #e4e4e4;border-top:8px solid #e4e4e4}.reflex-container.vertical>.reflex-splitter.reflex-thin{border-left:8px solid hsla(0,0%,100%,0);border-right:8px solid hsla(0,0%,100%,0);cursor:col-resize;height:100%;margin:0 -8px;width:17px!important}.reflex-container.vertical>.reflex-splitter.reflex-thin.active,.reflex-container.vertical>.reflex-splitter.reflex-thin:hover{border-left:8px solid #e4e4e4;border-right:8px solid #e4e4e4}
/*# sourceMappingURL=main.e3cdd0ec.css.map*/